#!/usr/bin/env python3
import itertools, math
import scipy.stats as stats
import sys


class FastAreader:
    """Read fasta file from specified fname or STDIN if not given"""

    def __init__(self, fname=""):
        """contructor: saves attribute fname"""

        self.fname = fname
        self.fileH = None

    def doOpen(self):
        if self.fname == "":
            return sys.stdin
        else:
            return open(self.fname)

    def readFasta(self):
        header = ""
        sequence = ""

        with self.doOpen() as self.fileH:
            header = ""
            sequence = ""

            # skip to first fasta header
            line = self.fileH.readline()
            while not line.startswith(">"):
                line = self.fileH.readline()
            header = line[1:].rstrip()

            for line in self.fileH:
                if line.startswith(">"):
                    yield header, sequence
                    header = line[1:].rstrip()
                    sequence = ""
                else:
                    sequence += "".join(line.rstrip().split()).upper()

        yield header, sequence


########################################################################
# CommandLine
########################################################################


class CommandLine:
    """
    Handle the command line, usage and help requests.

    attributes:
    all arguments received from the commandline using .add_argument will be
    avalable within the .args attribute of object instantiated from CommandLine.
    For example, if myCommandLine is an object of the class, and requiredbool was
    set as an option using add_argument, then myCommandLine.args.requiredbool will
    name that option.

    """

    def __init__(self, inOpts=None):
        """
        CommandLine constructor.
        Implements a parser to interpret the command line argv string using argparse.
        """

        import argparse

        self.parser = argparse.ArgumentParser(
            description="Program prolog - read a genome and return under-represented kMers",
            epilog="Thank you",
            add_help=True,  # default is True
            prefix_chars="-",
            usage="%(prog)s [options] -option1[default] <input >output",
        )

        self.parser.add_argument(
            "-m",
            "--max",
            nargs="?",
            default=8,
            type=int,
            action="store",
            help="max kMer size ",
        )
        self.parser.add_argument(
            "-c",
            "--cutoffProb",
            nargs="?",
            type=float,
            default=-4.0,
            action="store",
            #
            help="lowest Zscore allowed for printing ",
        )

        self.parser.add_argument(
            "-p",
            "--pseudoCount",
            type=int,
            default=1,
            action="store",
            help="pseudocount to be added to each count",
        )
        self.parser.add_argument(
            "-P",
            "--Palindrome",
            type=bool,
            nargs="?",
            const=True,
            default=False,
            action="store",
            help="set if only palindromes print",
        )

        self.parser.add_argument(
            "-v", "--version", action="version", version="%(prog)s 0.1"
        )
        if inOpts is None:
            self.args = self.parser.parse_args()
        else:
            self.args = self.parser.parse_args(inOpts)


def rc(s):
    """Return the reverse complement of the sequence."""
    return s[::-1].translate(
        str.maketrans("ACGTN", "TGCAN")
    )  # N is needed for composites


class Genome:
    bases = "ACGT"

    def __init__(self, min=1, max=8, pseudo=1):
        # initialize all kMers to the pseudocount value
        self.counts = {}
        for k in range(min, max + 1):  # kMer size
            newKcounts = {
                "".join(kmer): pseudo
                for kmer in itertools.product(Genome.bases, repeat=k)
            }
            self.counts.update(newKcounts)

        self.min, self.max, self.pseudo = min, max, pseudo
        self.n = sum(self.counts.values())  # add the pseudos to initialize n

    def addSequence(self, seq):
        """Count Kmers in the genome, and build composite sums that will be needs for odd kMers."""
        for l in range(self.min, self.max + 1):
            for p in range(len(seq) - self.max):
                try:
                    self.counts[seq[p : p + l]] += 1
                except KeyError:
                    pass
        self.n += len(seq) - self.max
        # build the composite sequences for all of the odd ones and their components, then put them in counts
        composites = {}
        for s in self.counts:
            if (len(s) > 2) and (
                len(s) % 2 == 1
            ):  # it's odd, so build a composite sum by adding over all center bases.
                flank = len(s) // 2
                newSeq = s[:flank] + "N" + s[flank + 1 :]  # build the composite
                composites[newSeq] = sum(
                    (
                        self.counts[s[:flank] + b + s[(flank + 1) :]]
                        for b in Genome.bases
                    )
                )
                # build left and right k-1mers and central k-2mer
                composites[newSeq[:-1]] = sum(
                    (
                        self.counts[s[:flank] + b + s[(flank + 1) : -1]]
                        for b in Genome.bases
                    )
                )
                composites[newSeq[1:]] = sum(
                    (
                        self.counts[s[1:flank] + b + s[(flank + 1) :]]
                        for b in Genome.bases
                    )
                )
                composites[newSeq[1:-1]] = sum(
                    (
                        self.counts[s[1:flank] + b + s[(flank + 1) : -1]]
                        for b in Genome.bases
                    )
                )
        self.counts.update(composites)

    def E(self, s):
        """return expected count of s. P(ab) = P(a|b) * P(b). P(abc) = P(a|b) * P(b|c) * P(c)"""
        # P(abcd) = P(a) * P(b|a) * P(c|ab) * P(d|bc)
        # E(abcd) = c(a) * c(ab)/c(a) * c(abc)/c(ab) * c(bcd)/c(bc) *len(s) = c(abc) * c(bcd) / c(bc)
        # for odd length s, assumes that the composites have already been built

        try:
            return self.counts[s[:-1]] * self.counts[s[1:]] / self.counts[s[1:-1]]
        except KeyError:
            return self.counts[s]
        except ZeroDivisionError:
            return 0.0

    def pValue(self, s):
        return stats.binom.cdf(self.counts[s], self.n, self.E(s) / self.n)

    def Zscore(self, s):
        """calculate Zscore
        s = kMer sequence
        """
        mu = self.E(s)
        var = mu * (1 - mu / self.n)  # using binomial, estimate variance
        if var == 0.0:
            return 1000000000
        return (self.counts[s] - mu) / math.sqrt(var)

    def Evalue(self, s):
        return self.pValue(s) * self.n


########################################################################
# Main
# Here is the main program
#
#
########################################################################


def main(myCommandLine=None):
    """
    Report underRepresented kMers in a genome.

    STDIN: FastA file of genome
    Options:
        -c cutoff value in Zscores. Sequences with lower Zscore probs will be reported [-4.0]
        -p pseudoCount value if desired [0]
        -P boolean flag specifying that only RC Palindromes should be reported [False]
        -m maximum kMer size to analyze.[8]

    :param myCommandLine:
    :return:
    """
    cl = CommandLine()

    # get the genome
    sourceReader = FastAreader()

    # setup the Tetramer object using centerSequences as the reference
    thisGenome = Genome(1, cl.args.max, cl.args.pseudoCount)
    molecules = 0
    for head, seq in sourceReader.readFasta():
        thisGenome.addSequence(seq)
        molecules += 1
    print("N = {}, molecules = {}".format(thisGenome.n, molecules))

    outArray = []
    for seq, count in thisGenome.counts.items():
        # for odd len sequence, we just print the composite that is for the full sequence.
        if "N" in seq:  # this is a composite
            p = seq.find("N")
            if len(seq) != (p * 2 + 1):
                continue  # skip this one, it is a part of a composite
        elif len(seq) % 2 == 1:
            continue  # odd sequences print only if they are a composite

        outArray.append(
            [
                seq,
                count,
                thisGenome.E(seq),
                thisGenome.Evalue(seq),
                thisGenome.Zscore(seq),
            ]
        )
    outArray.sort(key=lambda e: (len(e[0]), -e[4]), reverse=True)
    for seq, count, E, eVal, Z in outArray:
        if (seq in thisGenome.counts) and (Z <= cl.args.cutoffProb):
            # if the -P flag was set, only print palindromes, otherwise print palindromes with a flag
            rcFlag = ""
            if rc(seq) != seq:  # this must not be a palindrome
                if cl.args.Palindrome:
                    continue  # dont print non palindromes
            elif (
                not cl.args.Palindrome
            ):  # it is a palindrome so if we are printing everything give it a flag
                rcFlag = "RC palindrome"

            print(
                "{0:8}\t{1:0d}\t{2:0.2f}\t{3:0.2e}\t{4:0.2f}\t{5:13}".format(
                    seq, count, E, eVal, Z, rcFlag
                )
            )


__all__ = ["Genome", "rc"]

if __name__ == "__main__":
    main()



import chameleontools.FastAreader as FastAreader
import chameleontools.SeqParser as SeqParser
import Bio.SeqRecord
import random


class InvalidAA(KeyError):
    """
    Inherits from KeyError
    """
    def __init__(self, message) -> None:
        super().__init__(message)

class CDSanalyzer:
    """Counts up and calculates frequency of codons given a CDS

    Returns:
        CDSanalyzer: Contains counts and frequencies of codons from given CDS collection
    """
    dnaCodonTable = {
        # DNA codon table
        # T
        "TTT": "F",
        "TCT": "S",
        "TAT": "Y",
        "TGT": "C",  # TxT
        "TTC": "F",
        "TCC": "S",
        "TAC": "Y",
        "TGC": "C",  # TxC
        "TTA": "L",
        "TCA": "S",
        "TAA": "*",
        "TGA": "*",  # TxA
        "TTG": "L",
        "TCG": "S",
        "TAG": "*",
        "TGG": "W",  # TxG
        # C
        "CTT": "L",
        "CCT": "P",
        "CAT": "H",
        "CGT": "R",  # CxT
        "CTC": "L",
        "CCC": "P",
        "CAC": "H",
        "CGC": "R",  # CxC
        "CTA": "L",
        "CCA": "P",
        "CAA": "Q",
        "CGA": "R",  # CxA
        "CTG": "L",
        "CCG": "P",
        "CAG": "Q",
        "CGG": "R",  # CxG
        # A
        "ATT": "I",
        "ACT": "T",
        "AAT": "N",
        "AGT": "S",  # AxT
        "ATC": "I",
        "ACC": "T",
        "AAC": "N",
        "AGC": "S",  # AxC
        "ATA": "I",
        "ACA": "T",
        "AAA": "K",
        "AGA": "R",  # AxA
        "ATG": "M",
        "ACG": "T",
        "AAG": "K",
        "AGG": "R",  # AxG
        # G
        "GTT": "V",
        "GCT": "A",
        "GAT": "D",
        "GGT": "G",  # GxT
        "GTC": "V",
        "GCC": "A",
        "GAC": "D",
        "GGC": "G",  # GxC
        "GTA": "V",
        "GCA": "A",
        "GAA": "E",
        "GGA": "G",  # GxA
        "GTG": "V",
        "GCG": "A",
        "GAG": "E",
        "GGG": "G",  # GxG
    }

    validAA = set(dnaCodonTable.values())

    def __init__(
        self, input: str | SeqParser.StealthGenome | list[Bio.SeqRecord.SeqRecord] = None
    ) -> None:
        """Creates dictionaries to store and calculate codon counts and frequencies

        Args:
            input (str | SeqParser.StealthGenome | list[Bio.SeqRecord.SeqRecord]): FastA file or an iterable sample of CDS 
        """
        self.codonUsage = {
            "F": {"TTT": 0, "TTC": 0},
            "L": {"TTA": 0, "TTG": 0, "CTT": 0, "CTC": 0, "CTA": 0, "CTG": 0},
            "S": {"TCT": 0, "TCC": 0, "TCA": 0, "TCG": 0, "AGT": 0, "AGC": 0},
            "Y": {"TAT": 0, "TAC": 0},
            "C": {"TGT": 0, "TGC": 0},
            "W": {"TGG": 0},
            "P": {"CCT": 0, "CCC": 0, "CCA": 0, "CCG": 0},
            "H": {"CAT": 0, "CAC": 0},
            "Q": {"CAA": 0, "CAG": 0},
            "R": {"CGT": 0, "CGC": 0, "CGA": 0, "CGG": 0, "AGA": 0, "AGG": 0},
            "I": {"ATT": 0, "ATC": 0, "ATA": 0},
            "M": {"ATG": 0},
            "T": {"ACT": 0, "ACC": 0, "ACA": 0, "ACG": 0},
            "N": {"AAT": 0, "AAC": 0},
            "K": {"AAA": 0, "AAG": 0},
            "V": {"GTT": 0, "GTC": 0, "GTA": 0, "GTG": 0},
            "A": {"GCT": 0, "GCC": 0, "GCA": 0, "GCG": 0},
            "D": {"GAT": 0, "GAC": 0},
            "E": {"GAA": 0, "GAG": 0},
            "G": {"GGT": 0, "GGC": 0, "GGA": 0, "GGG": 0},
            "*": {"TAA": 0, "TAG": 0, "TGA": 0},
        }
        self.count = 0
        self.codonFreq = {k: {} for k in self.codonUsage.keys()}
        self._read_file(input) if type(input) == str else self._read_StealthGenome(
            input
        )
        self._calculate()

    def _read_file(self, input: str) -> None:
        """Helper method, reads in FastA file

        Args:
            input (str): FastA file
        """
        freader = FastAreader.FastAreader(input)
        for _, cds in freader.readFasta():
            cds = cds.seq
            for i in range(len(cds), 3):
                codon = input[i : i + 3]
                codon_to_aa = self.dnaCodonTable[codon]
                self.codonUsage[codon_to_aa][codon] += 1

    def _read_StealthGenome(
        self, input: SeqParser.StealthGenome | list[Bio.SeqRecord.SeqRecord]
    ) -> None:
        """Helper method, reads CDS from StealthGenome object or iterable of SeqRecord

        Args:
            input (SeqParser.StealthGenome | list[Bio.SeqRecord.SeqRecord]): Iterable SeqRecord of CDS
        """
        input = input.cds_sequence if type(input) == SeqParser.StealthGenome else input
        for cds in input:
            cds = str(cds.seq)
            for i in range(0, len(cds), 3):
                codon = cds[i : i + 3]
                codon_to_aa = self.dnaCodonTable[codon]
                self.codonUsage[codon_to_aa][codon] += 1

    def _calculate(self):
        """Helper method, calculates codon frequency from given codon counts
        """
        for aa, codon_dict in self.codonUsage.items():
            aa_sum = sum(codon_dict.values())
            self.count = +aa_sum
            for codon in codon_dict.keys():
                if aa_sum == 0:
                    self.codonFreq[aa][codon] = 1 / len(codon_dict)
                else:
                    self.codonFreq[aa][codon] = self.codonUsage[aa][codon] / aa_sum

    def _format_aa_arg(self, aa: str):
        """Helper method, type and input checking AA inputs

        Args:
            aa (str): Single Letter AA code

        Raises:
            TypeError: Non-string type is input
            InvalidAA: Invalid AA code is input

        Returns:
            str: formatted single letter AA
        """
        if type(aa) != str:
            raise TypeError(f"Expected {type(str())} got {type(aa)}")
        aa = aa.upper()
        if len(aa) != 1 or aa not in self.validAA:
            raise InvalidAA(f"Invalid single-letter Amino Acid: {aa}")
        return aa

    def addCDS(self, cds: str) -> None:
        """Adds input CDS to existing dictionary

        Args:
            cds (str): CDS sequence, in frame
        """
        for i in range(len(cds), 3):
            codon = cds[i : i + 3]
            codon_to_aa = self.dnaCodonTable[codon]
            self.codonUsage[codon_to_aa][codon] += 1
        self._calculate()

    def len(self) -> int:
        """Length

        Returns:
            int: NT count of all analyzes CDS 
        """
        return self.count

    def getFrequency(
        self, aa: str = None
    ) -> dict[str, dict[str, float]] | dict[str, float]:
        """Get the codon frequency table or codon frequency by AA

        Args:
            aa (str, optional): Single letter AA. Defaults to None.

        Returns:
            dict[str, dict[str, float]] | dict[str, float]: Entire codon frequency table or codon frequency by AA
        """
        if aa == None:
            return self.codonFreq
        aa = self._format_aa_arg(aa)
        return self.codonFreq[aa]

    def getUsage(self, aa: str = None) -> dict[str, dict[str, int]] | dict[str, int]:
        """Get the codon usage table or codon usage by AA

        Args:
            aa (str, optional): Single letter AA. Defaults to None.

        Returns:
            dict[str, dict[str, float]] | dict[str, float]: Entire codon usage table or codon usage by AA
        """
        if aa == None:
            return self.codonUsage
        aa = self._format_aa_arg(aa)
        return self.codonUsage[aa]

class CodonOptimizer:
    """
    This is a basic CodonOptimizer that takes codon usage data and regenerates sequences
    roughly fit to levels of tRNA availibility in an organism.
    """

    def __init__(self, frequencies: dict):
        """Initializes codon usage data

        Args:
            frequencies (dict): _description_
        """
        self.synFreqDic = frequencies

    def _weightedChoice(self, aa: str):
        """Generates a codon by weighted random given an AA

        Args:
            aa (str): Single letter AA 

        Returns:
            str: Weighted random codon from given AA
        """
        codons = list(self.synFreqDic[aa].keys())
        frequencies = self.synFreqDic[aa].values()
        chosenCodon = random.choices(codons, weights=frequencies, k=1)[0]
        return chosenCodon

    def assembleSeed(self, aaSequence: str):
        """Reverse translates a AA sequence based on codon usage statistics 

        Args:
            aaSequence (str): AA sequence to be reverse translated

        Returns:
            str: DNA coding sequence
        """
        
        dnaList = [self._weightedChoice(aa) for aa in aaSequence]
        dnaSeed = "".join(dnaList)
        return dnaSeed

    def getFrequency(self) -> dict[str, dict[str, float]]:
        """Gets the codon usage of object

        Returns:
            dict[str, dict[str, float]]: Codon frequency grouped by AA
        """
        return self.synFreqDic


__all__ = ["CDSanalyzer", "CodonOptimizer", "InvalidAA"]



from Bio import SeqIO
from Bio.SeqFeature import SimpleLocation
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from typing import Iterator, Literal
from chameleontools.ORFfinder import ORFfinder
from chameleontools.FastAreader import FastAreader
import sys


class StealthGenome:
    """
    Reads genome file in genbank or FASTA format,

    Stores Genome sequence and CDS regions
    """

    def _validArg(self, genome: str):
        """"Helper method, checks valid args"

        Args:
            genome (str): input file name

        Returns:
            bool: valid input
        """
        g_valid = False, False, False
        g_valid = (
            genome.endswith(".fasta")
            or genome.endswith(".fa")
            or genome.endswith(".gb")
            or genome.endswith(".gbk")
        )
        return g_valid

    def _fastaGenome(self, fasta: str):
        """Helper method, handles reading in of FastA files

        Args:
            fasta (str): FastA filename

        Returns:
            tuple[list[Bio.SeqRecord.SeqRecord],Bio.Seq.Seq]: List of CDS regions and genome sequence
        """
        reader = FastAreader(fasta)
        CDS_records = []
        sequence = []
        for frag, (_, seq) in enumerate(reader.readFasta()):
            genome = Seq(seq)
            orfs = ORFfinder(
                seq, longestGene=True, minGene=200
            ).get_genes()  # array of estimated genes
            for i, j in enumerate(orfs):
                "Converting ORFfinder output to Biopython SimpleLocation"
                loc = SimpleLocation(j[0] - 1, j[1], strand=1 if j[3] > 0 else -1)

                "Building Biopython SeqRecord for CDS"
                rec = SeqRecord(loc.extract(genome))
                rec.description = "Predicted Protein Coding Sequence"
                rec.name = "Predicted CDS HEADER"
                rec.id = f"FRAG_{frag} CDS_{i} for CODON STATS"

                CDS_records.append(rec)
            sequence.append(genome)
        return CDS_records, sequence

    ############################################################################################
    # ^^^ Private Helper Functions ^^^
    ############################################################################################

    def __init__(self, genome_infile: str) -> None:
        """Validates inputs and stores genome sequence and CDS regions

        Args:
            genome_infile (str): genome file in GenBank or FastA format
            
        Usage:

            self.getGenome() -> array of Seq()

            self.getCDS() -> List of SeqRecord

            self.filetype() -> type of input file | "fasta" if FastA file, "genbank" if GenBank file
        """
        if genome_infile != None:
            g_bool = self._validArg(genome_infile)

            if not g_bool:
                print(
                    f"Genome File needs to be FASTA file (.fasta || .fa) or GenBank file (.gb || .gbk). Got {genome_infile}",
                    file=sys.stderr,
                )
                exit()

            if genome_infile.endswith(".fasta") or genome_infile.endswith(".fa"):
                self.cds_sequence, self.genome_sequence = self._fastaGenome(
                    genome_infile
                )
                self.input = "fasta"
            else:
                self.cds_sequence,self.genome_sequence = self._Stealth_CDS_Extract(genome_infile)
                self.input = "genbank"

    def _Stealth_CDS_Extract(self, gbfile: str) -> tuple[list[Seq],Iterator[SeqRecord]]:
        """Generator for CDSs in a GenBank file
        
        Args:
            gbfile (str): input GenBank file
        """
        gb = SeqIO.parse(gbfile, "genbank") 
        for rec in gb:
            genome_sequence = [
                rec.seq
            ]  # define self.genome_sequence from genbank record
            def _generator() -> Iterator[SeqRecord]:
                count = 0
                for feature in rec.features:
                    if feature.type == "CDS" or feature.type == "ORF":
                        count += 1
                        out = feature.extract(rec)

                        """Set the description to 'product' qualifier if avaiable. Else use 'label' qualifier or [null] if labels are not present"""
                        out.description = feature.qualifiers.get(
                            "product", feature.qualifiers.get("label", ["null"])
                        )[0]

                        out.name = "TEMP CDS HEADER"
                        out.id = f"CDS_{count} CODON STATS"
                        yield out
            return  _generator(), genome_sequence  # Only accepts single entry genbank records

    def filetype(self) -> Literal["fasta", "genbank"]:
        """Returns filetype of input file

        Returns:
            Literal[str]: "fasta" or "genbank"
        """
        return self.input

    def getGenome(self) -> list[Seq]:
        """Returns genome sequence(s)

        Returns:
            list[Seq]: Sequence(s) of genome stored in a list
        """
        return self.genome_sequence

    def getCDS(self) -> list[SeqRecord]:
        """Returns iterable of CDS regions

        Returns:
            list[SeqRecord]: list of CDSs

        Yields:
            Iterator[SeqRecord]: Iterable of CDSs
        """
        return self.cds_sequence


class PlasmidParse:
    """Reads in an annotated plasmid file in genbank format,
    parses mutable regions from CDS/ORF annotations,
    takes statistics for total mutable range
    """

    def _validArg(self, plasmid_infile: str) -> bool:
        """Helper method, checks for valid input
        
        Args:
            plasmid_infile (str): filename for input file
            
        Returns:
            bool: boolean of valid filename """
        return plasmid_infile.endswith(".gb") or plasmid_infile.endswith(".gbk")

    def _parsePlasmid(
        self, gbfile: str
    ) -> tuple[list[SimpleLocation], list[SimpleLocation]]:
        """Helper func, filters for CDS and noncoding sequence

        Args:
            gbfile (str): input plasmid sequence

        Returns:
            tuple[list[SimpleLocation], list[SimpleLocation]]: regions to avoid and CDS regions
        """
        avoid, cds = set(),set() # sets to avoid duplicate annotation regions
        gb = SeqIO.parse(gbfile, "genbank")
        for rec in gb:
            self.record = rec
            for feature in rec.features:
                '''SimpleLocation unhashable, break down into data and recreate object'''
                simpLoc = feature.location
                simpLoc_data = (simpLoc.start,simpLoc.end,simpLoc.strand)
                feat_data = (simpLoc_data, simpLoc.start % 3) # keeps location, frame
                if feature.type in self.ignore_annotations:
                    continue
                if feature.type in self.keep_annotations: 
                    cds.add(feat_data)
                else:
                    avoid.add(feat_data)
            break  # Only handles single entry genbank records
        
        '''recreate SimpleLocation objects'''
        avoid = [(SimpleLocation(x[0],x[1],x[2]),y) for x,y in avoid]
        cds = [(SimpleLocation(x[0],x[1],x[2]),y) for x,y in cds] 
        return avoid, cds

    def _trimStart(
        self, loc_list: list[SimpleLocation], cds: SimpleLocation
    ) -> list[SimpleLocation]:
        """Helper method, trims the start codon of CDS from others that overlap, handles fwd/rev strands

        Args:
            loc_list (list[SimpleLocation]): list of SimpleLocation objects to subtract from
            cds (SimpleLocation): CDS to subtract

        Returns:
            list[SimpleLocation]: list of trimmed SimpleLocation objects
        """
        ref = (
            SimpleLocation(cds.start, cds.start + 3, cds.strand)
            if cds.strand > 0
            else SimpleLocation(cds.end - 3, cds.end, cds.strand)
        )
        for i in range(len(loc_list)):
            loc = loc_list[i]
            if loc.start < ref.start and loc.end > ref.end:
                loc_list[i] = SimpleLocation(loc.start, ref.start, loc.strand)
                loc_list.insert(i + 1, SimpleLocation(ref.end, loc.end, loc.strand))
                break
            if loc.start in ref:
                loc_list[i] = SimpleLocation(ref.end, loc.end, loc.strand)
                break
            elif loc.end in ref or loc.end == ref.end:
                loc_list[i] = SimpleLocation(loc.start, ref.start, loc.strand)
                break
        return

    def _frameAdj(self, loc_list: list[SimpleLocation], frame: int) -> list[SimpleLocation]:
        """Helper func, trims mutable regions to lie in frame

        Args:
            loc_list (list[SimpleLocation]): List of SimpleLocations to frame adjust
            frame (int): Frame to adjust to

        Returns:
            List[SimpleLocation]: List of frame-adjusted SimpleLocations
        """
        new = []
        for loc in loc_list:
            st = loc.start
            ed = loc.end
            for _ in range(1, 3):
                if st % 3 != frame:
                    st += 1
                if ed % 3 != frame:
                    ed -= 1
            if (ed - st) < 3:
                continue
            adj = SimpleLocation(st, ed, loc.strand)
            new.append(adj)
        return new

    def _removeOverlap(
        self, loc_list: list[SimpleLocation], cds: SimpleLocation
    ) -> list[SimpleLocation]:
        """Helper method, removes overlapping cds regions

        Args:
            loc_list (list[SimpleLocation]): List of locations to subtract from
            cds (SimpleLocation): CDS to subtract 

        Returns:
            list[SimpleLocation]: List of non-overlapping regions of CDS
        """
        ret = []
        for loc in loc_list:
            new = [[loc.start, loc.end]]
            if loc.start in cds and (loc.end in cds or loc.end == cds.end):
                "completely removes region"
                continue
            if loc.start < cds.start and loc.end > cds.end:
                "splits range into two"
                dummy = [cds.end, loc.end]
                new[-1][1] = cds.start
                new.append(dummy)
            elif loc.start in cds:
                "truncates start"
                new[-1][0] = cds.end
            elif loc.end in cds or loc.end == cds.end:
                "truncates end"
                new[-1][1] = cds.start + 1
            ret.extend([SimpleLocation(i, j, loc.strand) for i, j in new])
        return ret

    def _windowCorrection(
        self,
        old: list[list[SimpleLocation], int, SimpleLocation],
        st_bound: int,
        ed_bound: int,
    ) -> list[list[SimpleLocation], int, SimpleLocation]:
        """Helper method, corrects for out-frame overlapping CDS

        Args:
            old (list[list[SimpleLocation], int, SimpleLocation]): list of current mutable regions
            st_bound (int): beginning of overlapping CDS
            ed_bound (int): end of overlapping CDS

        Returns:
            list[list[SimpleLocation], int, SimpleLocation]: New list of mutable regions
        """
        loc, fm, parent = old
        new = SimpleLocation(st_bound, ed_bound, parent.strand)
        dummy = []

        for i in loc:
            if st_bound > i.end:
                continue
            if i.start in new and (i.end in new or i.end == new.end):
                dummy.append(i)
            elif i.start in new:
                dummy.append(SimpleLocation(i.start, ed_bound, i.strand))
                continue
            elif i.end in new or i.end == new.end:
                dummy.append(SimpleLocation(st_bound, i.end, i.strand))
                continue

        return [dummy, fm, new]

    ############################################################################################
    # Private Helper Functions
    ############################################################################################

    def __init__(self, plasmid_infile: str, keep_annotation = [], ignore_annotation = []) -> None:
        """Validates inputs, parses out mutable regions

        Args:
            plasmid_infiled (str): plasmid file in FASTA or genbank format

        Usage:
            self.getGenBank() -> Biopython SeqRecord() of input plasmid, contains feature annotations

            self.getSeq() -> Biopython Seq() of input Plasmid

            self.regions() -> array of Biopython SimpleLocation() ranges of mutable codons

            self.mutableCount() -> final mutable regions on plasmid

            self.unmutableCount() -> number of removed basepairs from total CDS coverage

            self.regionCount() -> Total CDS coverage before extracting mutable regions

        
        """
        p_bool = self._validArg(plasmid_infile)
        if not p_bool:
            print(
                f"Plasmid Input File needs to be GenBank file (.gb || .gbk). Got: {plasmid_infile}",
                file=sys.stderr,
            )
            exit()

        self.keep_annotations = {"CDS","ORF","gene"}
        self.ignore_annotations = {"source"}
        
        for annotation in keep_annotation:
            self.keep_annotations.add(annotation)
        for annotation in ignore_annotation:
            self.ignore_annotations.add(annotation)
        
        "Did this cause it was ugly when using tuple unpacking"
        temp_parse = self._parsePlasmid(plasmid_infile)

        temp = self.defineMutable(temp_parse)

        self.record = self.record

        self.plasmid_sequence = (
            self.record.seq
        )  # self.record -> Bio.SeqRecord of plasmid

        self.mutable_regions = temp[0]
        self.total_coverage = temp[1]
        self.removed = temp[2]

        self.total_mutable = self.total_coverage - self.removed

    def defineMutable(
        self, location: tuple[list[SimpleLocation]]
    ) -> tuple[list[SimpleLocation],int,int]:
        """
        Parses out all mutable regions based on plasmid annotations
        """
        
        
        "Finds non-overlapped mutable regions in an annotated plasmid"

        avoid, cdsRegions = location
        temp = []  # [[[SimpLocation List],frame,parent CDS]]

        "Removes mutable cds regions from overlapping non-coding sequence"
        for cds, frame in cdsRegions:
            if len(cds) <= 3:
                continue
            bounds = [[cds.start, cds.end]]
            for loc, _ in avoid:
                if cds.start > loc.start and cds.end < loc.end:
                    "case: interior ORF to non-coding region"
                    bounds = [[-1, -1]]
                    break
                if bounds[-1][0] < loc.start and bounds[-1][1] > loc.end:
                    "case: non-coding region interior to ORF"
                    new_bound = [loc.end, cds.end]
                    bounds[-1][1] = loc.start
                    bounds.append(new_bound)
                elif bounds[-1][0] in loc:
                    "case: clipped start"
                    bounds[-1][0] = loc.end
                elif bounds[-1][1] in loc or bounds[-1][1] == loc.end:
                    "case: clipped end"
                    bounds[-1][1] = loc.start
            if (bounds[-1][1] - bounds[-1][0]) <= 3:
                del bounds[-1]
            if len(bounds) <= 0:
                continue
            mut_range = [
                SimpleLocation(start=rng[0], end=rng[1], strand=cds.strand)
                for rng in bounds
            ]
            temp.append([mut_range, frame, cds])

        "Trimming start codons"
        for i in range(len(temp)):
            tLoc = temp[i][0]
            frm = temp[i][1]
            reg = temp[i][2]
            window = temp[i + 1 :]
            for j in range(len(window)):
                _, parent_frame, del_cds = window[j]
                if (del_cds.start in reg and del_cds.strand > 0) or (
                    del_cds.end in reg and del_cds.strand < 0
                ):
                    self._trimStart(temp[i][0], del_cds)
                    tLoc = temp[i][0]
            if reg.strand > 0:
                trimmed = SimpleLocation(reg.start + 3, reg.end, reg.strand)
                if tLoc[0].start == reg.start:
                    temp[i][0][0] = SimpleLocation(
                        trimmed.start, tLoc[0].end, tLoc[0].strand
                    )
            else:
                trimmed = SimpleLocation(reg.start, reg.end - 3, reg.strand)
                if tLoc[-1].end == reg.end:
                    temp[i][0][-1] = SimpleLocation(
                        tLoc[-1].start, trimmed.end, tLoc[-1].strand
                    )

        "sorts regions by CDS length, compares larger CDS to smaller CDS to remove overlap"
        temp = sorted(temp, key=lambda x: len(x[2]), reverse=True)

        "remove overlaps"
        for i in range(len(temp)):
            frag = temp[i][0]
            frm = temp[i][1]
            region = temp[i][2]
            window = temp[i + 1 :]
            if region == None:
                continue
            for j in range(len(window)):
                _, parent_frame, parent_cds = window[j]
                if parent_cds == None:
                    continue
                if parent_cds.start in region or parent_cds.end in region:
                    if parent_frame != frm or parent_cds.strand != region.strand:
                        temp[i][0] = self._removeOverlap(frag, parent_cds)
                        frag = temp[i][0]
                if parent_cds.start in region and parent_cds.end in region:
                    temp[i + j + 1][2] = None
                elif parent_cds.start in region:
                    temp[i + j + 1] = self._windowCorrection(
                        window[j], region.end, parent_cds.end
                    )
                elif parent_cds.end in region:
                    temp[i + j + 1] = self._windowCorrection(
                        window[j], parent_cds.start, region.start
                    )

        "frame adjusts mutable fragments to lie in frame of parent CDS"
        for i in range(len(temp)):
            temp[i][0] = self._frameAdj(temp[i][0], temp[i][1])

        mutable_regions = sorted(
            [region for CDS, _, tf in temp for region in CDS if tf is not None],
            key=lambda x: x.start,
        )

        "mutable region statistics - total covered region and mutable regions"
        total_coverage = sum([len(x[-1]) for x in temp if x[-1] is not None])
        removed = total_coverage - sum([len(x) for x in mutable_regions])
        return mutable_regions, total_coverage, removed

    def getSeq(self) -> Seq:
        """Get the sequence of the plasmid

        Returns:
            Seq: Plasmid NT sequence
        """
        return self.plasmid_sequence

    def regions(self) -> list[SimpleLocation]:
        """Get the list of mutable regions

        Returns:
            list[SimpleLocation]: SimpleLocations defining mutable regions
        """
        return self.mutable_regions

    def mutableCount(self) -> int:
        """Get length of mutable regions

        Returns:
            int: NT count of mutable regions
        """
        return self.total_mutable

    def regionCount(self) -> int:
        """Get total length of CDS regions

        Returns:
            int: NT count of CDS regions
        """
        return self.total_coverage

    def unmutableCount(self) -> int:
        """Get length of unremoved regions

        Returns:
            int: NT count of unmutable CDS regions
        """
        return self.removed

    def getGenBank(self) -> SeqRecord:
        """Get GenBank record of plasmid infile

        Returns:
            SeqRecord: SeqRecord of input plasmid
        """
        return self.record


__all__ = ["StealthGenome", "PlasmidParse"]



import chameleontools.SeqParser as SeqParser
import chameleontools.StealthParser as StealthParser
import chameleontools.Stealth as Stealth
import chameleontools.CodonAnalyzer as CodonAnalyzer
import chameleontools.ChromatoSeq as ChromatoSeq
import Bio.Seq as Seq, Bio.SeqFeature as SeqFeature
import Bio.SeqIO as SeqIO
import sys
from tqdm import tqdm


class pipeline:
    def _openOutfile(self, outfile: str):
        if outfile == None:
            return sys.stdout
        else:
            return open(outfile, "w")

    def __init__(
        self,
        genome_infile: str,
        plasmid_infile: str,
        outfile: str,
        z_score: float,
        pseudo: int,
        kMax: int,
        kMin: int,
        palindrome: bool = False,
        keep: list[str] = [],
        ignore: list[str] = []
    ) -> None:
        self.loading_bar = tqdm(total=100, desc="Reading Files")  # CLI loading bar

        self.stealth_input = SeqParser.StealthGenome(genome_infile)
        self.plasmid_input = SeqParser.PlasmidParse(plasmid_infile, keep_annotation= keep, ignore_annotation= ignore)

        self.loading_bar.update(15)

        self.stealth_motifs = self.load_stealth(z_score, pseudo, kMax, kMin, palindrome)

        self.unremoved_motifs = 0

        self.total_motifs = 0

        self.run(outfile)

    def load_stealth(
        self,
        z_score: float,
        pseudo: int,
        kMax: int,
        kMin: int,
        palindrome: bool = False,
    ):
        self.loading_bar.set_description("Performing Stealth Analysis")

        stealth_out = Stealth.StealthV0(self.stealth_input, z_score, pseudo, kMax, kMin)
        self.loading_bar.update(10)
        return StealthParser.ParseStealth(stealth_out, palindrome)

    def codon_analysis(self) -> CodonAnalyzer.CDSanalyzer:
        self.loading_bar.set_description("Gathering Codon Statistics")

        codon_usage = CodonAnalyzer.CDSanalyzer(self.stealth_input)
        self.loading_bar.update(20)
        return codon_usage

    def hide_motif(self, region: list[SeqFeature.SimpleLocation]):
        plasmid_seq = self.plasmid_input.getSeq()
        codon_freq = self.codon_analysis().getFrequency()
        codon_optimizer = CodonAnalyzer.CodonOptimizer(codon_freq)
        pattern_remover = ChromatoSeq.patternConstrainer(
            self.stealth_motifs, codon_freq
        )
        motif_check = ChromatoSeq.MotifChecker(self.stealth_motifs)

        def _hide_motif_helper(aaSeq: str) -> Seq.Seq:
            roundsData = []
            for _ in range(5):
                seed = codon_optimizer.assembleSeed(aaSeq)
                optimization = pattern_remover.optimizeSequence(seed)

                mortality = motif_check.checkMotifs(optimization)
                roundsData.append((mortality, optimization))
            lowest_mortality = min(roundsData, key=lambda x: x[0])
            self.unremoved_motifs += lowest_mortality[0]
            best_seq = lowest_mortality[1]
            return Seq.Seq(best_seq)

        self.loading_bar.set_description("Removing Motifs")

        modified_blocks = []  # list of SimpleLocations, mapped to modified Seq()
        
        loading_progress = 48 / (sum([len(x) for x in region])) # more dynamic loading bar for motif removal
        
        for s_loc in region:
            coding_seq = s_loc.extract(plasmid_seq)
            self.total_motifs += motif_check.checkMotifs(coding_seq)
            aa_seq = coding_seq.translate()
            stealth_seq = _hide_motif_helper(aa_seq)
            if s_loc.strand == -1:
                stealth_seq = stealth_seq.reverse_complement()
            modified_blocks.append([s_loc, stealth_seq])

            self.loading_bar.update((loading_progress * len(s_loc)))

        return modified_blocks

    def run(self, outfile: str):
        mutable_regions = self.plasmid_input.regions()
        blocks = self.hide_motif(mutable_regions)
        plasmid_sequence = self.plasmid_input.getSeq()
        new_seq = Seq.Seq("")
        cut_start = 0

        self.loading_bar.set_description("Assembling Stealth Plasmid")

        for s_loc, seq in blocks:
            new_seq += plasmid_sequence[cut_start : s_loc.start]
            new_seq += seq
            cut_start = s_loc.end

            self.loading_bar.update((7 / len(blocks)))

        new_seq += plasmid_sequence[cut_start:]

        stealth_record = self.plasmid_input.getGenBank()
        stealth_record.seq = new_seq  # Overwrite old sequence with Stealth sequence

        with self._openOutfile(outfile) as fd:
            SeqIO.write(stealth_record, fd, "genbank")

        self.loading_bar.close()

    def writeOut(self):
        total_motifs = self.total_motifs
        remaining_motifs = self.unremoved_motifs
        removed_motifs = total_motifs - remaining_motifs

        plasmid_name = self.plasmid_input.getGenBank().name
        
        total_cds = self.plasmid_input.regionCount()
        mutable = self.plasmid_input.mutableCount()

        print(
            f"""{plasmid_name}: {mutable} NT were mutable from {total_cds} NT of CDS coverge\nRemoved {removed_motifs} of {total_motifs} motifs from {plasmid_name}. {remaining_motifs} motifs remain.""",
            file=sys.stderr,
        )


import argparse


def main():
    parser = argparse.ArgumentParser(
        description="Stealth optimizes an input plasmid sequence to a given organism's genome",
        usage=f"""pstealth --genome (-g) <genome infile> --plasmid (-p) <plasmid infile> --outfile -o [outfile | default: stdout] [-zPMmrs]
            Optional Args:
                --zScore (-z) -> [zscore cutoff value | default: -4]
                --pseudo (-P) [pseudo-count value | default: 0]
                --max (-M) [maximum motif size | default: 8]
                --min (-m) [minimum motif size | default: 1]
                --palindrome (-r) [Remove RC palindromes only | default: off]
                --silent (-s) [Hide report message | default: show]
                --keep (-k) [Adds annotations to consider in Mutable Regions | default = ('ORF','gene','CDS')]
                --ignore (-i) [Adds annotations to ignore when defining Mutable Regions | default = ('source')]""",
    )
    parser.add_argument(
        "-g",
        "--genome",
        type=str,
        action="store",
        help="input genome file in GenBank or FastA format",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--outfile",
        default=None,
        type=str,
        action="store",
        help="output filename, default = STDOUT",
    )
    parser.add_argument(
        "-p",
        "--plasmid",
        type=str,
        action="store",
        help="input plasmid file in GenBank format",
        required=True,
    )
    parser.add_argument(
        "-z",
        "--zScore",
        type=float,
        default=-4,
        action="store",
        help="set Z-score cutoff for underrepresented motifs, default = -4",
    )
    parser.add_argument(
        "-P",
        "--pseudo",
        type=int,
        default=0,
        action="store",
        help="set pseudo-count value, default = 0",
    )
    parser.add_argument(
        "-M",
        "--max",
        type= int,
        default=8,
        choices=range(2, 10),
        help="Set maximum motif size, default = 8",
    )
    parser.add_argument(
        "-m",
        "--min",
        type=int,
        default=1,
        choices=range(1, 9),
        help="Set minimum motif size, default = 1",
    )
    parser.add_argument(
        "-r",
        "--palindrome",
        default=False,
        action="store_true",
        help="Hide only RC palindromes, default = off",
    )
    parser.add_argument(
        "-s",
        "--silent",
        default=False,
        action="store_true",
        help="Disable final modification report, default = show",
    )
    
    parser.add_argument(
        "-k",
        "--keep",
        type= str,
        default= None,
        action="store",
        nargs = '+',
        help="Add plasmid annotations to what are considered mutable regions- case sensitive, Default = {'ORF','gene','CDS'}",
    )
    
    parser.add_argument(
        "-i",
        "--ignore",
        type= str,
        default= None,
        action="store",
        nargs = '+',
        help="Add plasmid annotations to ignore when defining mutable regions- case sensitive, Default = {'source'}",
    )

    args = parser.parse_args()

    genome_infile = args.genome
    plasmid_infile = args.plasmid
    outfile = args.outfile
    z_score = args.zScore
    pseudo = args.pseudo
    kMax = args.max
    kMin = args.min
    palindrome = args.palindrome
    keep_annotation = args.keep if args.keep is not None else []
    ignore_annotation = args.ignore if args.ignore is not None else []
    
    dummy = pipeline(
        genome_infile, plasmid_infile, outfile, z_score, pseudo, kMax, kMin, palindrome, keep_annotation, ignore_annotation
    )

    if not args.silent:
        dummy.writeOut()




class ORFfinder:
    """ORF finder class that scans through a genome and reports ORFs from all frames
    """

    def __init__(
        self,
        seq: str,
        longestGene=False,
        minGene=100,
        starts: set = {"ATG"},
        stops: set = {"TAA", "TAG", "TGA"},
    ):
        """Finds all potential ORFs in a given sequence

        Args:
            seq (str): genomic sequence
            longestGene (bool, optional): Toggles finding largest gene per ORF. Defaults to False.
            minGene (int, optional): Gene length report cutoff. Defaults to 100.
            starts (set, optional): Define start codons. Defaults to {"ATG"}.
            stops (set, optional): Define stop codons. Defaults to {"TAA", "TAG", "TGA"}.
        """
        assert minGene >= 0
        self.geneCandidates = []
        format = seq.strip().upper().replace(" ", "")
        flipTable = format[::-1].maketrans(
            "ATCG", "TAGC"
        )  # Builds reverse compliment strand

        self.fiveThree = format
        self.threeFive = format[::-1].translate(flipTable)
        self.validStart = starts
        self.validStops = stops
        self.minGene = minGene
        self.longestGene = longestGene

        self._analysis()

    def _geneFinder(self, seq, rev, longGene):
        """Helper function, collects starts and identifies stops
        """
        start = set()
        for frame in range(0, 3):
            start.clear()
            for i in range(frame, len(seq), 3):
                codon = seq[i : i + 3]

                if codon in self.validStart:
                    start.add(i)

                if codon in self.validStops and len(start) > 0:
                    _frame = -1 - frame if rev else frame + 1
                    self._addGene(start, i, _frame, longGene)
                    start = set()

            if len(start) > 0:
                _frame = -1 - frame if rev else frame + 1
                self._addGene(start, len(seq) - 3, _frame, longGene)

    def _addGene(self, start, i, frame, longGene):
        """Helper Function, does all gene-candidate calculations 
        """
        for s in sorted(start):
            length = i - s + 3
            if frame < 0:  # Handles reverse strand calcs
                startPos = len(self.fiveThree) - i - 2
                stopPos = len(self.fiveThree) - s
            else:  # Handles top strand
                startPos = s + 1
                stopPos = i + 3
            if length >= self.minGene:
                self.geneCandidates.append([startPos, stopPos, length, frame])
            if longGene:  # LongestGene flag
                return

    def _analysis(self):
        """Runs ORF finder on input sequence
        """
        self._geneFinder(self.fiveThree, False, self.longestGene)
        self._geneFinder(self.threeFive, True, self.longestGene)

        # sorts geneCandidates list, sorting by gene length
        self.geneCandidates = [
            x
            for x in sorted(
                self.geneCandidates, key=lambda entry: entry[2], reverse=True
            )
            if x[2] % 3 == 0 ]

    def get_genes(self) -> list[tuple[int, int, int, int]]:
        """Fetches all gene candidates
        
        Returns:
            list[tuple[int, int, int, int]]: List of gene candidates
        """
        return self.geneCandidates


__all__ = ["ORFfinder"]



"""
TABI - UCSC iGEM 2023

Classes to parse Stealth output and isolate restricted motifs

Author: Tyler Gaw
"""

# Takes in a file path to Stealth output and generates a set of expanded degenerate motifs
# If no file_path is specified, reads from STDIN
# Usage Example:
#     motif_set_1 = ParseStealth(file_path_1)
#     motif_set_2 = PalindromeParseStealth(file_path_2)
#     conserved_motifs = motif_set_1.intersect(motif_set_2) -> set

from ._StealthParser import *



"""
TABI - UCSC iGEM 2023

Classes to parse Stealth output and isolate restricted motifs

Author: Tyler Gaw
"""

# Takes in a file path to Stealth output and generates a set of expanded degenerate motifs
# If no file_path is specified, reads from STDIN
# Usage Example:
#     motif_set_1 = ParseStealth(file_path_1)
#     motif_set_2 = PalindromeParseStealth(file_path_2)
#     conserved_motifs = motif_set_1.intersect(motif_set_2) -> set

from typing import Iterator
from chameleontools.Stealth import StealthV0
import sys


class ParseStealth(set):
    """
    Takes stealth output data as a text file and processes a set of motifs.
    """

    degenerateBases = {  # Degenerate base pair dictionary
        "R": ["A", "G"],
        "Y": ["C", "T"],
        "M": ["A", "C"],
        "K": ["G", "T"],
        "S": ["C", "G"],
        "W": ["A", "T"],
        "H": ["A", "C", "T"],
        "B": ["C", "G", "T"],
        "V": ["A", "C", "G"],
        "D": ["A", "G", "T"],
        "N": ["A", "C", "G", "T"],
    }

    def __init__(
        self, file_path: str | StealthV0 = sys.stdin, palindrome: bool = False
    ):
        """
        Constructor: Inherits from set\n
        args : <file_path> - File path to stealth output\n
        Output: Set containing expanded motifs
        """
        super().__init__()
        if isinstance(file_path, StealthV0):
            self._readStealth(file_path, palindrome)
        else:
            self._readIn(file_path, palindrome)

    def _readStealth(self, obj: StealthV0, pal: bool) -> set:
        for seq, _, t_f in obj.getOutput():
            if pal:
                if t_f:
                    for expanded_motif in self._permute(seq):
                        if self._isRevComp(expanded_motif):
                            self.add(expanded_motif)
            else:
                for expanded_motif in self._permute(seq):
                    self.add(expanded_motif)

    def _readIn(self, file_path: str, pal: bool) -> set:
        """
        helper func - from file reads all motifs into set
        """
        file_p = open(file_path, "r") if type(file_path) == str else file_path
        with file_p as fd:
            while line := fd.readline().strip():
                if line.startswith("N ="):
                    continue
                line = line.split()
                if pal:
                    if line[-1] == "Palindrome":
                        for expanded_motif in self._permute(line[0]):
                            if self._isRevComp(expanded_motif):
                                self.add(expanded_motif)
                else:
                    for expanded_motif in self._permute(line[0]):
                        self.add(expanded_motif)

    def _permute(self, string, cur_idx=0, cur_perm="") -> Iterator[str]:
        """
        helper func - generates permutations of all degenerate sequences
        """
        if cur_idx == len(string):
            yield cur_perm
            if cur_perm == string:
                return
        else:
            cur_let = string[cur_idx]
            if cur_let in self.degenerateBases:
                subs = self.degenerateBases[cur_let]
                for sub in subs:
                    yield from self._permute(string, cur_idx + 1, cur_perm + sub)
            else:
                yield from self._permute(string, cur_idx + 1, cur_perm + cur_let)

    def _isRevComp(self, seq: str) -> bool:
        """
        helper function - boolean check on reverse compliment
        """
        t = {"A": "T", "C": "G", "T": "A", "G": "C"}
        if len(seq) % 2 == 0:
            return "".join([t[i] for i in seq[::-1]]) == seq
        else:
            return (
                "".join([t[i] for i in seq[: len(seq) // 2 : -1]])
                == seq[: len(seq) // 2]
            )


__all__ = ["ParseStealth"]



#!/usr/bin/env python3
"""
TABI - UCSC iGEM 2023

Commandline interface for parsing Stealth output

Standalone program for ParseStealth.py

Author: Tyler Gaw
"""

# Usage: ./pipeline.py -i <input file | default= stdin> -o <outfile | default stdout> -[optional]
#         -s [sorted | default= False]
#         -p [RC paindromes only | default= False]
#         -d DNAworks compatible output


import argparse, sys, re
from chameleontools.StealthParser import ParseStealth


def writeFile(motif: set, outfile: str, sortBool: bool, header: str, dnaWorks: bool):
    """
    writes stealth output to DNAworks format file
    """
    file_p = open(outfile, "w") if type(outfile) == str else sys.stdout
    with file_p as fd:
        fd.write(f"Run Options: {header}\n")
        if sortBool:
            "explicit sort by length first, then alphabetical"
            for i, seq in enumerate(sorted(list(motif), key=lambda x: (len(x), x))):
                descriptor = f" re{i}" if dnaWorks else f" [temp]"
                fd.write(f"{descriptor}\t{seq}\n")
        else:
            for seq in list(motif):
                descriptor = f" re{i}" if dnaWorks else f" [temp]"
                fd.write(f"{descriptor}\t{seq}\n")


def main():
    """
    Commandline parse, executes file write
    """
    script_name = re.split(r"\\|/", sys.argv[0])[-1]
    parser = argparse.ArgumentParser(
        description="Reads in Stealth file, outputs motifs in formatted file",
        usage=f"{script_name} -i <input file> -o <outfile | default stdout> -[optional]\n\t-s [sorted | default= False]\n\t-p [RC paindromes only | default= False]\n\t-d DNAworks compatible output",
    )
    parser.add_argument(
        "--infile", "-i", type=str, action="store", help="input file", required=True
    )
    parser.add_argument(
        "--outfile",
        "-o",
        default=sys.stdout,
        type=str,
        action="store",
        help="output file",
    )
    parser.add_argument(
        "--sorted", "-s", default=False, action="store_true", help="sort output"
    )
    parser.add_argument(
        "--palindrome",
        "-p",
        default=False,
        action="store_true",
        help="palindrome output",
    )
    parser.add_argument(
        "--dnaWorks",
        "-d",
        default=False,
        action="store_true",
        help="DNAWorks compatible output",
    )
    args = parser.parse_args()
    conserved = ParseStealth(args.infile, args.palindrome)
    header = f"{script_name} {' '.join(sys.argv[1:])}"
    writeFile(conserved, args.outfile, args.sorted, header, args.dnaWorks)
    # import os
    # parser.add_argument("--file",'-f', nargs="?", default=".", help="Path to file or directory")
    # args = parser.parse_args()

    # # Get the absolute path of the file or directory specified
    # target_path = os.path.abspath(args.file)

    # # Access the working directory
    # current_directory = os.getcwd()

    # print(f"target path {target_path}, CWD {current_directory}")



from typing import Iterator
import sys


class FastAreader:
    """
    Define objects to read FastA files.

    Args:
    fname (str): file name (optional), default is None (STDIN)

    Usage:
    thisReader = FastAreader ('testTiny.fa')
    for head, seq in thisReader.readFasta():
        print (head,seq)
    """

    def __init__(self, fname=None):
        """
        Constructor: saves attribute fname
        Args:
        fname (str): file name (optional), default is None (STDIN)
        """
        self.fname = fname

    def doOpen(self):
        """
        Handle file opens, allowing STDIN.

        Returns:
        file: either sys.stdin or file handler for the input file
        """
        if self.fname is None:
            return sys.stdin
        else:
            return open(self.fname)

    def readFasta(self) -> Iterator[tuple[str, str]]:
        """
        Read an entire FastA record and return the sequence header/sequence

        Yields:
        tuple: header and sequence as a tuple
        """
        header = ""
        sequence = ""

        with self.doOpen() as fileH:
            header = ""
            sequence = ""

            # skip to first fasta header
            line = fileH.readline()
            while not line.startswith(">"):
                line = fileH.readline()
            header = line[1:].rstrip()

            for line in fileH:
                if line.startswith(">"):
                    yield header, sequence
                    header = line[1:].rstrip()
                    sequence = ""
                else:
                    sequence += line.strip()
        yield header, sequence


__all__ = ["FastAreader"]



"""
TABI - UCSC iGEM 2023

Module: CodonAnalyzer

Description: Classes to calculate codon usage statistics based on a collection of CDS from an organism

Author: Tyler Gaw, James Larbalestier
"""

from ._CodonAnalyzer import *



from Bio.Seq import Seq

class patternConstrainer:
    """
    Initialized via an instance of putativly "optimized" sequence,
    patternConstrainer finds regions of the seed that contain motifs of
    interest (via a 5'->3' sliding window approach) and passes associated
    indicies (start pos. based on specific window size & fixed adapter
    sequences at the begainning and end of each window) to some outside
    func.
    In this case, the outside func returns a 'more ideal' (synonomous variate)
    which patternConstrainer catches and overwrites into the working sequence.
    """

    def __init__(self, motifs, codonPref):
        self.motifs = motifs  # A list of motifs to avoid, likely to trigger RMS
        self.codonOptions = { # The dictionary of codons mapping to their corresponding amino acids.
            "F": ["TTT", "TTC"],
            "L": ["TTA", "TTG", "CTT", "CTC", "CTA", "CTG"],
            "S": ["TCT", "TCC", "TCA", "TCG", "AGT", "AGC"],
            "Y": ["TAT", "TAC"],
            "C": ["TGT", "TGC"],
            "W": ["TGG"],
            "P": ["CCT", "CCC", "CCA", "CCG"],
            "H": ["CAT", "CAC"],
            "Q": ["CAA", "CAG"],
            "R": ["CGT", "CGC", "CGA", "CGG", "AGA", "AGG"],
            "I": ["ATT", "ATC", "ATA"],
            "M": ["ATG"],
            "T": ["ACT", "ACC", "ACA", "ACG"],
            "N": ["AAT", "AAC"],
            "K": ["AAA", "AAG"],
            "V": ["GTT", "GTC", "GTA", "GTG"],
            "A": ["GCT", "GCC", "GCA", "GCG"],
            "D": ["GAT", "GAC"],
            "E": ["GAA", "GAG"],
            "G": ["GGT", "GGC", "GGA", "GGG"],
            "*": ["TAA", "TAG", "TGA"],
        } 
        self.codonPref = codonPref

    def motifInSeq(self, sequence):
        """
        Given a sequence and initialized with a list of motifs, evaluates presence or absence of motifs in the sequence.
        In practice this method is used to support the sliding window approach to pattern constraint requirements.
        """
        for motif in self.motifs:
            if motif in sequence:
                return True  # Return True immediately. This indicates that the sequence contains a motif that we want to avoid.
        return False  # If none of the motifs were found, return False. This indicates that the sequence is safe to use as is.

    def optimizeCodon(self, sequence, start):
        '''
        Organizes optimization of the codon sequence at a given start position in a DNA sequence.

        This method takes a DNA sequence and a start position as input. It extracts a 12-base pair
        (12-bp) subsequence starting from the given position and translates it into an amino acid sequence.
        The method then calculates head and tail subsequences based on the start position and a fixed size
        window. The head and tail subsequences are used to integrate the current window of the sliding
        window approach with the rest of the sequence.

        The method employs a Permutations object to rank the 12-bp DNA sequence based on motifs and codon preferences.
        The resulting cleaned region is then inserted back into the original DNA sequence, replacing the original
        12-bp sequence. The final sequence with the optimized codon is returned.

        We chose a 12-bp/ 4 codon window as the longest motif we are screening is 8-bp long, and to resolve that the 
        window must have at least a 4 codon scope. The reason that the window is not larger is to retain stochastic
        reconstruction as much as possible, and to reduce the footprint of the pattern constraint optimization.
        '''
        length = len(sequence)
        workingDnaSequence = sequence[start : start + 12]

        bioSeq = Seq(workingDnaSequence)
        workingAaSeq = bioSeq.translate()

        if start < 9:
            headStart = start
        elif 9 <= start:
            headStart = start - 9

        if length <= start + 21:
            tailEnd = length
            tailStart = length
        else:
            tailEnd = start + 21
            tailStart = start + 12

        headAdapt = sequence[
            headStart:start
        ]  # change to handle begainning case (no -6 index)
        tailAdapt = sequence[tailStart:tailEnd]

        myPermuter = Permutations(self.motifs, self.codonPref)
        cleanedRegion = myPermuter.rank(
            workingDnaSequence, workingAaSeq, headAdapt, tailAdapt
        )

        cleanCutOff = len(cleanedRegion) - len(tailAdapt)
        newSequence = (
            sequence[:start]
            + cleanedRegion[len(headAdapt) : cleanCutOff]
            + sequence[start + 12 :]
        )

        return newSequence

    def optimizeSequence(self, sequence) -> str:
        """
        This method iterates through the input DNA sequence, looking for motifs within 12-base pair (12-bp) subsequences.
        When a motif is found, the method calls the 'optimizeCodon' method to optimize the codon at the motif's start position.
        The optimized DNA sequence is updated, and the process continues until all relevant motifs have been optimized.
        """
        optimizedSequence = sequence
        for i in range(9, len(sequence) - 9, 3):
            if self.motifInSeq(optimizedSequence[i : i + 12]):
                newSequence = self.optimizeCodon(optimizedSequence, i)
                optimizedSequence = newSequence
        return optimizedSequence

class Permutations:
    """
    Generates all possible codon permutations of given window (codon subsequence
    of sequence) & ranks permutations based on rank items as they are generated.
    Each round of generation ends with the appension of scores and permutations
    added to net list as a tuple. These tuples are then sorted via score descrimination-
    only canditates with the best score for first item contenue, same with second item
    and then the 3rd item varies. This can be adjusted but allows for the prioritization
    of sorting methodologies for each specific score items. 
    """
    
    def __init__(self, motifs, frequencies):
        self.Motifs = motifs
        self.frequencies = {}
        for _,codon_dict in frequencies.items():
            self.frequencies.update(codon_dict)
        self.codonOptions = { # The dictionary of codons mapping to their corresponding amino acids.
            "F": ["TTT", "TTC"],
            "L": ["TTA", "TTG", "CTT", "CTC", "CTA", "CTG"],
            "S": ["TCT", "TCC", "TCA", "TCG", "AGT", "AGC"],
            "Y": ["TAT", "TAC"],
            "C": ["TGT", "TGC"],
            "W": ["TGG"],
            "P": ["CCT", "CCC", "CCA", "CCG"],
            "H": ["CAT", "CAC"],
            "Q": ["CAA", "CAG"],
            "R": ["CGT", "CGC", "CGA", "CGG", "AGA", "AGG"],
            "I": ["ATT", "ATC", "ATA"],
            "M": ["ATG"],
            "T": ["ACT", "ACC", "ACA", "ACG"],
            "N": ["AAT", "AAC"],
            "K": ["AAA", "AAG"],
            "V": ["GTT", "GTC", "GTA", "GTG"],
            "A": ["GCT", "GCC", "GCA", "GCG"],
            "D": ["GAT", "GAC"],
            "E": ["GAA", "GAG"],
            "G": ["GGT", "GGC", "GGA", "GGG"],
            "*": ["TAA", "TAG", "TGA"],
        }
        

    def permute(self, amino_acid_sequence, current_codons=""):
        '''
        Recusive generator to produce all possible codon permutations for a given amino acid sequence. 
        '''
        if not amino_acid_sequence:
            yield current_codons
            return

        current_aa = amino_acid_sequence[0]
        remaining_aas = amino_acid_sequence[1:]

        for codon in self.codonOptions[current_aa]:
            yield from self.permute(remaining_aas, current_codons + codon)

    def rank(self, dnaSeq, aaSeq, headAdapt, tailAdapt):
        """ 
        Organizes evalutation of each permutation (putative solution) for motifs present, stochasticity, & preferential 
        codons to select optimal solutions. This method uses the head & tail adapter sequences to consider cases where 
        motifs are split by the sliding window.
        """
        ranks = []
        permutation_generator = self.permute(aaSeq)

        for putative in permutation_generator:
            permutation = headAdapt + putative + tailAdapt

            severity, codScore = self.analyze_dna_motifs(permutation)
            adherance = self.gaugeAdherance(
                dnaSeq, putative
            )  # returns a value between 0 and 1, 0 being less adherant

            ranks.append((severity, adherance, codScore, permutation))

        min_first_item = min(ranks, key=lambda x: x[0])[0]
        filter1 = sorted(
            [t for t in ranks if t[0] == min_first_item],
            key=lambda x: x[1],
            reverse=True,
        )  # Sort the remaining tuples based on the second item in descending order

        max_second_item = filter1[0][1]
        filter2 = [
            t for t in filter1 if t[1] == max_second_item
        ]  # Extract tuples with the highest second item values

        max_third_item = max(filter2, key=lambda x: x[2])[2]
        filter3a = [
            t for t in filter2 if 0.95 * max_third_item <= t[2] <= 1.05 * max_third_item
        ]  # Extract tuples with a third item value within 5% of the maximal value
        filter3b = sorted(filter3a, key=lambda x: x[2], reverse=True)

        # add another layer, for now first item
        chosenTup = filter3b[0]
        chosenOne = chosenTup[3]
        return chosenOne

    def analyze_dna_motifs(self, sequence):
        '''
        Supports rank(), this method returns score values for motifs present & preferential 
        codon useage to elute optimal solutions.
        '''
        occurrences = {}  # Dictionary to store occurrences of short motifs
        total_occurrences = 0  # Total count of occurrences in the longer sequence

        for short_motif in self.Motifs:
            motif_positions = []  # List to store positions of motif occurrences
            motif_count = 0  # Count of occurrences for each motif
            start_position = 0

            while start_position < len(sequence):
                position = sequence.find(short_motif, start_position)

                if position == -1:
                    break  # No more occurrences found
                motif_positions.append(position)
                motif_count += 1
                total_occurrences += 1
                start_position = position + 1  # Move to the next position

            occurrences[short_motif] = {
                "positions": motif_positions,
                "count": motif_count,
            }

        score = 0.0
        codLen = len(sequence) / 3
        for i in range(0, len(sequence), 3):
            codon = sequence[i : i + 3]
            if codon in self.frequencies:
                score += self.frequencies[codon]
                Score = score / codLen
        return total_occurrences, Score

    def gaugeAdherance(self, refSeq, perm):
        '''
        Supports rank by providing score value reflective of stochasticity (origional input is stochastic).
        '''
        codonRef = []
        codonPerm = []
        adherance = 0
        
        for i in range(0, len(refSeq), 3):
            refCodon = refSeq[i : i + 3]
            codonRef.append(refCodon)

            perCodon = perm[i : i + 3]
            codonPerm.append(perCodon)

        adherance = 0
        codLen = len(refSeq) / 3
        for tup in zip(codonRef, codonPerm):
            cod1 = tup[0]
            cod2 = tup[1]
            if cod1 == cod2:
                adherance += 1 / codLen
        return adherance  # confers percentage of conserved codons in perm

class MotifChecker:
    """
    helper class, used to fetch final stats on motifs present in main
    """

    def __init__(self, motifs):
        self.motifs = motifs

    def checkMotifs(self, sequence):
        totalOccurrences = 0  # Total count of occurrences in the longer sequence

        for motif in self.motifs:
            startPosition = 0
            while startPosition < len(sequence):
                position = sequence.find(motif, startPosition)

                if position == -1:
                    break  # No more occurrences found

                totalOccurrences += 1
                startPosition = position + 1  # Move to the next position

        return totalOccurrences

__all__ = ["patternConstrainer","Permutations","MotifChecker"]



from ._stealth import Genome, rc as reverse_complement
from chameleontools.SeqParser import StealthGenome
from Bio.Seq import Seq


class StealthV0:
    def __init__(
        self,
        seq: Seq | str | list[str | Seq] | StealthGenome,
        zScore_cut: float,
        pseudo: int,
        kMax: int,
        kMin: int = 1,
    ) -> None:
        if type(seq) == StealthGenome:
            seq = seq.getGenome()
        self.output = self._runStealth(seq, zScore_cut, pseudo, kMax, kMin)

    def _runStealth(
        self, genome: str, zScore_cut: float, pseudo: int, kMax: int, kMin: int = 1
    ) -> list[tuple[str, float, bool]]:
        """
        runs Stealth analysis on input genome
        saves output to array -> [(motif,zScore)]
        """
        if type(genome) != list:
            genome = [genome]
        stealth = Genome(1, kMax, pseudo)
        for seq in genome:
            stealth.addSequence(str(seq))  # conversion of BioPython Seq() to str
        """Taken from original stealth.vo.py main() func"""
        ######V Written by David L. Bernick V#######
        outArray = []
        for seq, _ in stealth.counts.items():
            if len(seq) < kMin:
                continue
            # for odd  len sequence, we just print the composite that is for the full sequence.
            if "N" in seq:  # this is a composite
                p = seq.find("N")
                if len(seq) != (p * 2 + 1):
                    continue  # skip this one, it is a part of a composite
            elif len(seq) % 2 == 1:
                continue  # odd sequences print only if they are a composite
            ######^ Written by David L. Bernick ^#######
            zscore = stealth.Zscore(seq)
            if zscore <= zScore_cut:
                if seq in stealth.counts:
                    rc_flag = False
                    if reverse_complement(seq):
                        rc_flag = True
                    outArray.append((seq, zscore, rc_flag))
        return outArray

    def getOutput(self) -> list[tuple[str, float, bool]]:
        return self.output


__all__ = ["StealthV0"]
